"""
Wrapper around PyMef to read Mef3 files


=====================================================
Copyright 2022, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

PyMef by Jan Cimbalnick, Daniel Crepeau et al.

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from .IeegDataReader import IeegDataReader
from pymef.mef_session import MefSession


class Mef3Reader(IeegDataReader):

    mef_session = None
    mef_data = None

    def __init__(self, data_path, data_preload=False):
        super().__init__(data_path, data_preload)
        self.data_format = 'mef3'

        # read the session metadata
        try:
            self.mef_session = MefSession(self.data_path, '', read_metadata=True)
        except RuntimeError:
            logging.error('PyMef could not read data, either a password is needed or the data is corrupt')
            raise RuntimeError('PyMef could not read data')

        # TODO: check if sampling_rate and num_samples is equal for each channel

        # retrieve the sample-rate, total number of samples and channel names
        self.sampling_rate = self.mef_session.session_md['time_series_metadata']['section_2']['sampling_frequency'].item(0)
        self.num_samples = self.mef_session.session_md['time_series_metadata']['section_2']['number_of_samples'].item(0)
        self.channel_names = []
        for ts_channel_name, ts_channel_metadata in self.mef_session.session_md["time_series_channels"].items():
            self.channel_names.append(ts_channel_name)

        # (optionally) preload data
        if self.data_preload:

            # initialize empty array
            self.mef_data = []

            # loop over the channels
            for ts_channel_name, ts_channel_metadata in self.mef_session.session_md["time_series_channels"].items():

                # load the channel data
                try:
                    channel_data = self.mef_session.read_ts_channels_sample([ts_channel_name], (None, None))
                except Exception:
                    logging.error('PyMef could not read data, either a password is needed or the data is corrupt')
                    raise RuntimeError('Could not read data')

                # return and apply a conversion factor if needed
                channel_conversion_factor = ts_channel_metadata['section_2']['units_conversion_factor'].item(0)
                if channel_conversion_factor != 0 and channel_conversion_factor != 1:
                    channel_data[0] *= channel_conversion_factor

                # TODO: check MEF3 format, if after units_conversion_factor the units should be in uV, or after conversion
                #       just match whatever is in the Units Description

                # add channel
                self.mef_data.append(channel_data[0])


    def close(self):
        del self.mef_session
        if self.mef_data is not None:
            del self.mef_data



    #
    #
    #


    @staticmethod
    def __retrieve_channel_metadata(mef_session, channel_name):
        """
        Find and retrieve the MEF3 channel metadata and channel index by channel name
        """

        channel_metadata = None
        channel_counter = 0
        for ts_channel_name, ts_channel_metadata in mef_session.session_md["time_series_channels"].items():
            if ts_channel_name == channel_name:
                channel_metadata = ts_channel_metadata
                break
            channel_counter += 1

        if channel_metadata is None:
            logging.error('Could not find metadata for channel ' + channel_name + ', assuming there is no such channel in the dataset')
            return None, None

        return channel_metadata, channel_counter


    def retrieve_channel_data(self, channel_name, ensure_own_data=True):
        """
        Retrieve the MEF3 channel data by channel name

        Args:
            channel_name:                   The name of the channel for which to retrieve the data
            ensure_own_data (bool):         Should ensure the return a numpy array has it's own data (is not a view)

        Returns:
            Numpy array with data

        Raises:
            LookupError:                    Raised when the channel name cannot be found
            RuntimeError:                   Raised when unable to retrieve channel data
        """

        # find the channel metadata by channel name
        channel_metadata, channel_index = Mef3Reader.__retrieve_channel_metadata(self.mef_session, channel_name)
        if channel_metadata is None:
            raise LookupError('Could not find channel')

        # determine whether the data is preloaded
        if self.mef_data is None:
            # data is not preloaded

            # load the channel data
            try:
                channel_data = self.mef_session.read_ts_channels_sample([channel_name], (None, None))
            except Exception:
                logging.error('PyMef could not read data, either a password is needed or the data is corrupt')
                raise RuntimeError('Could not read data')

            # return and apply a conversion factor if needed
            channel_conversion_factor = channel_metadata['section_2']['units_conversion_factor'].item(0)

            if channel_conversion_factor != 0 and channel_conversion_factor != 1:
                channel_data[0] *= channel_conversion_factor
            return channel_data[0]

        else:
            # data is preloaded

            # return the channel data
            if ensure_own_data:
                return self.mef_data[channel_index].copy()
            else:
                return self.mef_data[channel_index]


    def retrieve_sample_range_data(self, channels, sample_start, sample_end, ensure_own_data=True):
        """
        Retrieve a specific range of MEF3 data for the requested channels

        Args:
            channels (str, list or tuple):  The channel(s) for which to retrieve the data.
                                            If empty, all channels will be retrieved
            sample_start (int):             The start-point in time (in samples) to start reading from (0-based)
            sample_end (int):               The sample to end the reading at (0-based)
            ensure_own_data (bool):         Should ensure the return a numpy array has it's own data (is not a view)

        Returns:
            data                            A float data matrix containing the signal data (of the requested channels)
                                            formatted as <channels> x <samples/time>

        Raises:
            LookupError:                    Raised when a channel name cannot be found
            RuntimeError:                   Raised when unable to retrieve data
        """

        # check/prepare the channels argument
        if isinstance(channels, str):
            channels = [channels]
        if isinstance(channels, tuple):
            channels = list(channels)
        if channels is None or len(channels) == 0:
            channels = self.channel_names

        if self.mef_data is None:
            # data is not preloaded

            # load the trial data
            try:
                sample_data = self.mef_session.read_ts_channels_sample(channels, [sample_start, sample_end])
                if sample_data is None or (len(sample_data) > 0 and sample_data[0] is None):
                    raise RuntimeError('Could not read data')

                # loop over the channels to retrieve
                for channel_counter in range(len(channels)):

                    # find the channel metadata by channel name
                    channel_metadata, _ = Mef3Reader.__retrieve_channel_metadata(self.mef_session, channels[channel_counter])
                    if channel_metadata is None:
                        raise LookupError('Could not find channel')

                    # apply a conversion factor if needed
                    channel_conversion_factor = channel_metadata['section_2']['units_conversion_factor'].item(0)
                    if channel_conversion_factor != 0 and channel_conversion_factor != 1:
                        sample_data[channel_counter] *= channel_conversion_factor

            except Exception:
                logging.error('PyMef could not read data, either a password is needed or the data is corrupt')
                raise RuntimeError('Could not read data')
        else:
            # data is preloaded

            # create a list with the numpy arrays
            sample_data = [None] * len(channels)

            # loop over the channels to retrieve
            for counter in range(len(channels)):

                # retrieve the index of the channel
                try:
                    channel_index = IeegDataReader.__retrieve_channel_metadata_mef(self.mef_session, self.channel_names.index(channels[counter]))
                except ValueError:
                    raise LookupError('Could not find channel')

                # pick the slice
                if ensure_own_data:
                    sample_data[counter] = self.mef_data[channel_index][sample_start:sample_end].copy()
                else:
                    sample_data[counter] = self.mef_data[channel_index][sample_start:sample_end]

        #
        return sample_data



"""
Functions to load BIDS sidecar metadata


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import logging
import pandas as pd


def load_channel_info(filepath):
    """

    Retrieve the channel metadata from a _channels.tsv file

    Args:
        filepath (str):             The path to the _channels.tsv file to load

    Returns:
        csv (dataframe):            A pandas dataframe containing the channels information

    Raises:
        FileNotFoundError:          If the file could not be found
        LookupError:                If one of the required columns could not be found

    """

    try:
        csv = pd.read_csv(filepath, sep='\t', header=0, encoding='unicode_escape', na_filter=False, dtype=str)
    except FileNotFoundError:
        logging.error('Could not find the file \'' + filepath + '\'')
        raise FileNotFoundError('Could not find file')

    # check the existence of required columns
    if 'name' not in csv.columns:
        logging.error('Could not find the \'name\' column in \'' + filepath + '\'')
        raise LookupError('Could not find column')
    if 'type' not in csv.columns:
        logging.error('Could not find the \'type\' column in \'' + filepath + '\'')
        raise LookupError('Could not find column')
    if 'status' not in csv.columns:
        logging.error('Could not find the \'status\' column in \'' + filepath + '\'')
        raise LookupError('Could not find column')

    #
    return csv


def load_event_info(filepath, addition_required_columns=None):
    """
    Retrieve the events from a _events.tsv file

    Args:
        filepath (str):                         The path to the _events.tsv file to load
        addition_required_columns(list/tuple):  One or multiple additional columns that need to be present in the _events.tsv

    Returns:
        csv (dataframe):                        A pandas dataframe containing the evetns information

    Raises:
        FileNotFoundError:                      If the file could not be found
        LookupError:                            If the mandatory 'onset' column or any of the required additional
                                                columns could not be found
    """

    try:
        csv = pd.read_csv(filepath, sep='\t', header=0, encoding='unicode_escape', na_filter=False, dtype=str)
    except FileNotFoundError:
        logging.error('Could not find the file \'' + filepath + '\'')
        raise FileNotFoundError('Could not find file')

    # check the existence of required columns
    if 'onset' not in csv.columns:
        logging.error('Could not find the \'onset\' column in \'' + filepath + '\'')
        raise LookupError('Could not find column')
    if addition_required_columns is not None:
        for column in addition_required_columns:
            if column not in csv.columns:
                logging.error('Could not find the \'' + column + '\' column in \'' + filepath + '\'')
                raise LookupError('Could not find column')

    #
    return csv


def load_ieeg_sidecar(filepath):
    """
    Read a JSON sidecar file

    Args:
        filepath (str):             The path to the JSON sidecar file to load

    Returns:
        ieeg_json (dict):           A dictionary containing the sidecar information

    Raises:
        IOError:                    If the file could not be found or accessed
        RuntimeError:               If the JSON file could not be parsed
    """

    # try to read the JSON configuration file
    try:
        with open(filepath) as json_file:
            ieeg_json = json.load(json_file)
    except IOError:
        logging.error('Could not access the IEEG JSON sidecar file at \'' + filepath + '\'')
        raise IOError('Could not access the IEEG JSON sidecar file')
    except json.decoder.JSONDecodeError as e:
        logging.error('Could not interpret the IEEG JSON sidecar file at \'' + filepath + '\', make sure the JSON syntax is valid: \'' + str(e) + '\'')
        raise RuntimeError('Could not interpret the IEEG JSON sidecar file')

    #
    return ieeg_json




"""
Helper class for the re-referencing of BIDS data


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from ieegprep.utils.misc import is_number


class RerefStruct:
    """
    Sets up and stores a re-referencing structure
    """

    groups = list()                     # re-referencing groups, each group entry holds the channels (names) that should be re-referenced over
    channel_group = dict()              # for each channel, the re-reference group that should be used to re-referencing it
    channel_exclude_epochs = None       # a dictionary with for each channel the epochs that should be excluded from re-referencing

    def __init__(self, groups, channel_reref_group):
        self.groups = groups
        self.channel_group = channel_reref_group

    @classmethod
    def generate_car(cls, channels):
        """
        Factory method to generate a Common Average Re-referencing (CAR) setup struct

        Args:
            channels (list or tuple):  Channels that might need to be re-referenced (or are needed for re-referencing)

        """

        # create a single group with all the channels from the channels argument
        groups = list()
        groups.append(channels.copy())

        # set each channel to be referenced to the same group
        channel_reref_group = dict()
        for channel in channels:
            channel_reref_group[channel] = 0

        return cls(groups, channel_reref_group)

    @classmethod
    def generate_car_per_headbox(cls, channel_names, channel_headboxes):
        """
        Factory method to generate a Common Average Re-referencing (CAR) per headbox setup struct

        Args:
            channel_names (list or tuple):      Channel names that might need to be re-referenced (or are needed for re-referencing)
            channel_headboxes (list or tuple):  Channel headboxes, order should correspond with the channels_names input argument

        """

        # create a single group with all the channels from the channels argument
        groups = list()
        channel_reref_group = dict()

        # loop over the unique headboxes
        unique_headboxes = sorted(list(set(channel_headboxes)))
        group_counter = 0
        for headbox in unique_headboxes:
            if is_number(headbox) and not headbox.lower() == 'nan' and not headbox.lower() == 'n/a':

                # find the channel names that belong to this headbox
                headbox_channels = [channel_names[ind] for ind, x in enumerate(channel_headboxes) if x == headbox]

                # store the channel names as a group
                groups.append(headbox_channels)

                # store for each channel in this group that they below to this group
                for channel_name in headbox_channels:
                    channel_reref_group[channel_name] = group_counter

                # raise the group counter
                group_counter += 1

        return cls(groups, channel_reref_group)


    def get_required_channels(self, retrieve_channels):
        """
        Lookup all the channels (names) that are needed for re-referencing, given the channels that need to be retrieved

        Args:
            retrieve_channels (list or tuple):  Channels that of which the data needs to be retrieved and re-referenced

        Note:   The list of channels that is returned can be different from the channels to need to be referenced (e.g. maybe
                only ECOG channels need to be re-referenced while re-referencing does have to occur over both ECOG and SEEG channels)
        """

        # create a list for all the re-references channels that are required for the channels that need to be retrieved
        all_channels = list()

        # loop over all requested channels
        for channel in retrieve_channels:
            if channel not in self.channel_group.keys():
                logging.error('Could not find requested channel ' + channel + ' in reref struct, make sure each included channel is also set for re-referencing')
                raise ValueError('Could not find requested channel')

            else:

                # for the requested channel, loop over the channels in the group that it needs for re-referencing
                for grp_channel in self.groups[self.channel_group[channel]]:

                    # add to list of required channels
                    if grp_channel not in all_channels:
                        all_channels.append(grp_channel)

        return all_channels


    def get_required_groups(self, retrieve_channels):
        """
        Retrieve all the groups that are needed for re-referencing, given the channels that need to be retrieved

        Args:
            retrieve_channels (list or tuple):  Channels that of which the data needs to be retrieved and re-referenced

        Note:   These can be different from the channels to need to be referenced (e.g. maybe only ECOG channels need
                to be re-referenced while re-referencing does have to occur over both ECOG and SEEG channels)
        """

        # create a list with all the groups required
        all_groups = list()

        # loop over all requested channels
        for channel in retrieve_channels:
            if channel not in self.channel_group.keys():
                logging.error('Could not find requested channel ' + channel + ' in reref struct')
                raise ValueError('Could not find requested channel')

            else:

                # add the group to the list of required groups
                if self.channel_group[channel] not in all_groups:
                    all_groups.append(self.channel_group[channel])

        return all_groups

    def set_exclude_reref_epochs(self, exclude_onsets, exclude_epoch=(-.01, 1.0), split_channel_names=None):
        """
        Set channel-epochs that should be excluded from re-referencing
        (this can be used when electrical stimulation was performed on specific channels at specific moments in the data)

        Args:
            exclude_onsets (dict with lists):   The onsets of the channel-epochs that need to be excluded from
                                                re-referencing. The argument should be a dictionary where the key
                                                of each entry in the dictionary represents the channel-name of which
                                                epochs should be excluded, and the value of the entry a list of
                                                onsets (in time) that needs to be excluded for that channel.
                                                A window around each onset - as defined by epoch - will define what
                                                will be excluded from re-referencing
            exclude_epoch (tuple):              The time-span that will be excluded around each onset in the
                                                channel-data. Expressed as a tuple with the start- and end-point in
                                                seconds relative to the onset (e.g. the standard tuple of '-1, 3' will
                                                exclude the signal in the period from 1s before the onset to 3s after
                                                onset).
            split_channel_names (str):          The channel-names to exclude ('exclude_onsets argument' keys) can actually
                                                consist of two channel names (which is the case for stimulated electrode
                                                pairs; e.g. a key could be 'Ch01-Ch02', where actually the onsets to
                                                exclude apply to both channel Ch01 and Ch02). If set, this arguments
                                                indicates on which character to split the channel-names (key) before
                                                adding the onsets to each of the separate channel-names to exclude.
        """

        # for each channel store the epoch windows which should be excluded from re-referencing
        self.channel_exclude_epochs = dict()
        for channel, onsets in exclude_onsets.items():

            if split_channel_names is None:
                # exclude_onsets dictionary keys are simply the channel-names

                self.channel_exclude_epochs[channel] = list()
                for onset in onsets:
                    self.channel_exclude_epochs[channel].append((onset + exclude_epoch[0], onset + exclude_epoch[1]))

            else:
                # exclude_onsets dictionary keys indicate multiple channel-names

                # separate the key into the separate channel-names
                split_channels = channel.split(split_channel_names)
                for sub_channel in split_channels:

                    # make sure the (sub) channel-name exists to hold exclusion epoch windows
                    if sub_channel not in self.channel_exclude_epochs.keys():
                        self.channel_exclude_epochs[sub_channel] = list()

                    for onset in onsets:
                        self.channel_exclude_epochs[sub_channel].append((onset + exclude_epoch[0], onset + exclude_epoch[1]))




"""
Functions to format and color console text

=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging


class ConsoleColors:
    """
    Class to hold console color code constants
    """

    RESET = '\033[0m'
    BOLD = '\033[1m'
    DISABLE = '\033[2m'
    UNDERLINE = '\033[4m'
    REVERSE = '\033[7m'
    INVISIBLE = '\033[8m'
    STRIKETHROUGH = '\033[9m'

    class FG:
        BLACK = '\033[30m'
        RED = '\033[31m'
        GREEN = '\033[32m'
        ORANGE = '\033[33m'
        BLUE = '\033[34m'
        PURPLE = '\033[35m'
        CYAN = '\033[36m'
        LIGHT_GREY = '\033[37m'
        DARKGREY = '\033[90m'
        LIGHT_RED = '\033[91m'
        LIGHT_GREEN = '\033[92m'
        YELLOW = '\033[93m'         # yellow = "\033[0;33;21m"
        LIGHT_BLUE = '\033[94m'
        PINK = '\033[95m'
        LIGHT_CYAN = '\033[96m'

    class BG:
        BLACK = '\033[40m'
        RED = '\033[41m'
        GREEN = '\033[42m'
        ORANGE = '\033[43m'
        BLUE = '\033[44m'
        PURPLE = '\033[45m'
        CYAN = '\033[46m'
        LIGHT_GREY = '\033[47m'

    @staticmethod
    def print_color(color, text):
        print(color + text + ConsoleColors.RESET)

    @staticmethod
    def print_green(text):
        print(ConsoleColors.FG.GREEN + text + ConsoleColors.RESET)

    @staticmethod
    def print_error(text):
        print(ConsoleColors.FG.RED + text + ConsoleColors.RESET)


class CustomLoggingFormatter(logging.Formatter):
    """
    A logging formatter to automatically color errors, warnings, info and debug output text
    """

    FORMATS = {
        logging.ERROR: ConsoleColors.FG.RED + "ERROR: %(filename)s: %(message)s" + ConsoleColors.RESET,
        logging.WARNING: ConsoleColors.FG.RED + "WARNING: %(message)s" + ConsoleColors.RESET,
        logging.INFO: "%(message)s",                     # TODO: currently no explicit color set, breaks 'print_progressbar'
        logging.DEBUG: "DEBUG: %(message)s",
        "DEFAULT": "%(message)s"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno, self.FORMATS['DEFAULT'])
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def multi_line_list(input_array, indent_length=45, first_line_caption='', items_per_line=4, item_delimiter=' ', first_line_single_item=None, no_item_text='-'):
    """
    Convert an array of strings (list or tuple) to a formatted string where the items are
    equally spread over multiple lines, instead of one long line

    Args:
        input_array (arroy of str):    String to format as text
        indent_length (int):           The spacing in front of the listed items
        first_line_caption (str):      The text in the first line that precedes the listing of the items in the array.
        items_per_line (int):          The number of items per line
        item_delimiter (str):          The str that is added between the array-items in the text
        first_line_single_item (str):  If set, will display this string on the first line, the rest of the items will
                                       then be shown from the second line on
        no_item_text (str):            Text if there are no items

    """
    current_line = ''
    return_text = ''

    if len(input_array) == 0:
        return first_line_caption.ljust(indent_length, ' ') + no_item_text

    #
    if first_line_single_item is not None:
        return_text = first_line_caption.ljust(indent_length, ' ') + str(first_line_single_item)
    else:
        current_line = first_line_caption

    # generate and print the lines
    sub_line = ''
    for i in range(len(input_array)):
        if not len(sub_line) == 0:
            sub_line += item_delimiter
        sub_line += str(input_array[i])
        if (i + 1) % items_per_line == 0:
            if not len(return_text) == 0:
                return_text += '\n'
            return_text += current_line.ljust(indent_length, ' ') + sub_line
            sub_line = ''
            current_line = ''

    # print the remaining items (if there are any)
    if not len(sub_line) == 0:
        if not len(return_text) == 0:
            return_text += '\n'
        return_text += current_line.ljust(indent_length, ' ') + sub_line

    return return_text


def print_progressbar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='█', print_end="\r"):
    """
    Call in a loop to create terminal progress bar

    Args:
        iteration (int):    current iteration (Int)
        total (int):        total iterations (Int)
        prefix (str):       prefix string (Str)
        suffix (str)        suffix string (Str)
        decimals (int)      positive number of decimals in percent complete (Int)
        length (int):       character length of bar (Int)
        fill (str):         bar fill character (Str)
        print_end (str):    end character (e.g. "\r", "\r\n") (Str)
        color (str):        text color (as ASCII code)

        Source: https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + '-' * (length - filled_length)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end=print_end)
    if iteration == total:
        print()




"""
Functions to load and epoch BIDS data
=====================================================


Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


def load_data_epochs(data_path, retrieve_channels, onsets,
                     trial_epoch=(-1, 3), baseline_norm=None, baseline_epoch=(-1, -0.1),
                     out_of_bound_handling='error',
                     high_pass=False, early_reref=None, line_noise_removal=None, late_reref=None, preproc_priority='mem'):
    """
    Load and epoch the data into a matrix based on channels, the trial onsets and the epoch range (relative to the onsets)

    Args:
        data_path (str):                    Path to the data file or folder
        retrieve_channels (list or tuple):  The channels (by name) of which the data should be retrieved, the output will
                                            be sorted accordingly according to this input argument.
        onsets (1d list or tuple):          The onsets of the trials around which to epoch the data
        trial_epoch (tuple):                The time-span that will be considered as the signal belonging to a single trial.
                                            Expressed as a tuple with the start- and end-point in seconds relative to
                                            the onset of the trial (e.g. the standard tuple of '-1, 3' will extract
                                            the signal in the period from 1s before the trial onset to 3s after trial onset).
        baseline_norm (None or str):        Baseline normalization setting [None, 'Mean' or 'Median']. If other than None,
                                            normalizes each trial epoch by subtracting the mean or median of part of the
                                            trial (the epoch of the trial indicated in baseline_epoch)
        baseline_epoch (tuple):             The time-span on which the baseline is calculated, expressed as a tuple with the
                                            start- and end-point in seconds relative to the trial onset (e.g. the
                                            standard tuple of '-1, -.1' will use the period from 1s before trial onset
                                            to 100ms before trial onset to calculate the baseline on); this argument
                                            is only used when baseline_norm is set to mean or median
        out_of_bound_handling (str):        Configure the handling of out-of-bound trial epochs;
                                                'error': (default) Throw an error and return when any epoch is out of bound;
                                                'first_last_only': Allows only the first trial epoch to start before the
                                                                   data-set and the last trial epoch to end beyond the
                                                                   length of the data-set, the trial epochs will be padded
                                                                   with NaN values. Note that the first and last trial are
                                                                   determined by the first and last entry in the 'onsets'
                                                                   parameter, which is not sorted by this function;
                                                'allow':           Allow trial epochs to be out-of-bound, NaNs values will
                                                                   be used for part of, or the entire, the trial epoch
        preproc_priority (str):             Set the preprocessing priority, can be set to either 'mem' (default) or 'speed'

    Returns:
        sampling_rate (int or double):      the sampling rate at which the data was acquired
        data (ndarray):                     A three-dimensional array with data epochs per channel (format: channel x
                                            trials/epochs x time); or None when an error occurs

    Note: this function's input arguments are in seconds relative to the trial onsets because the sample rate will
          only be known till after we read the data
    """
    pass




def load_data_epochs_averages(data_path, retrieve_channels, conditions_onsets,
                              trial_epoch=(-1, 3), baseline_norm=None, baseline_epoch=(-1, -0.1),
                              out_of_bound_handling='error', metric_callbacks=None,
                              high_pass=False, early_reref=None, line_noise_removal=None, late_reref=None, preproc_priority='mem'):


    """
    Load, epoch and return the average for each channel and condition (i.e. the signal in time averaged
    over all trials that belong to the same condition).

    Note: Because this function only has to return the average signal for each channel and condition, it is much more
          memory efficient (this is particularly important when the amount of memory is limited by a Docker or VM)
    Note 2: For the same reason, metric callbacks are implemented here, so while loading, but before averaging, metrics
            can be calculated on subsets of data with minimum memory usage. If memory is not an issue,
            load_data_epochs function can be used to retrieve the full dataset first and then perform calculations.

    Args:
        data_path (str):                      Path to the data file or folder
        retrieve_channels (list or tuple):    The channels (by name) of which the data should be retrieved, the output
                                              will be sorted accordingly
        conditions_onsets (2d list or tuple): A two-dimensional list to indicate the conditions, and the onsets
                                              of the trials that belong to each condition.
                                              (format: conditions x condition onsets)
        trial_epoch (tuple):                  The time-span that will be considered as the signal belonging to a single
                                              trial. Expressed as a tuple with the start- and end-point in seconds
                                              relative to onset of the trial (e.g. the standard tuple of '-1, 3' will
                                              extract the signal in the period from 1s before the trial onset to 3s
                                              after the trial onset).
        baseline_norm (None or str):          Baseline normalization setting [None, 'Mean' or 'Median']. If other
                                              than None, normalizes each trial epoch by subtracting the mean or median
                                              of part of the trial (the epoch of the trial indicated in baseline_epoch)
        baseline_epoch (tuple):               The time-span on which the baseline is calculated, expressed as a tuple with
                                              the start- and end-point in seconds relative to the trial onset (e.g. the
                                              standard tuple of '-1, -.1' will use the period from 1s before the trial
                                              onset to 100ms before the trial onset to calculate the baseline on);
                                              this argument is only used when baseline_norm is set to mean or median
        out_of_bound_handling (str):          Configure the handling of out-of-bound trial epochs;
                                                'error': (default) Throw an error and return when any epoch is out of bound;
                                                'first_last_only': Allows only the first trial epoch to start before the
                                                                   data-set and the last trial epoch to end beyond the
                                                                   length of the data-set, the trial epochs will be padded
                                                                   with NaN values. Note that the first and last trial are
                                                                   determined by the first and last entry in the 'onsets'
                                                                   parameter, which is not sorted by this function;
                                                'allow':           Allow trial epochs to be out-of-bound, NaNs values will
                                                                   be used for part of, or the entire, the trial epoch
        metric_callbacks (func or tuple):     Function or tuple of functions that are called to calculate metrics based
                                              on subsets of the un-averaged data. The function(s) are called per
                                              with the following input arguments:
                                                sampling_rate -    The sampling rate of the data
                                                data -             A subset of the data in a 2d array: trials x samples
                                                baseline -         The corresponding baseline values for the trials
                                              If callbacks are defined, a third variable is returned that holds the
                                              return values of the metric callbacks in the format: channel x condition x metric
        preproc_priority (str):              Set the preprocessing priority, can be set to either 'mem' (default) or 'speed'

    Returns:
        sampling_rate (int or double):        The sampling rate at which the data was acquired
        data (ndarray):                       A three-dimensional array with signal averages per channel and condition
                                              (format: channel x condition x samples); or None when an error occurs
        metrics (ndarray):                    If metric callbacks are specified, will return a three-dimensional array
                                              with the metric callback results (format: channel x condition x metric),
                                              else wise None

    Note: this function input arguments are in seconds relative to trial onsets because the sample rate will
          only be known till after we read the data
    """
    pass

